import dataclasses
import uuid
from typing import List, Literal, Optional, Union

from .common import GenericPath
from .validation import ValidationException

FileOnCloseBehavior = Literal["none", "rm-if-finished"]


@dataclasses.dataclass
class StdioDef:
    """
    If `path` is `None`, then the default HQ path will be used.
    """

    path: Optional[GenericPath]
    on_close: FileOnCloseBehavior

    @staticmethod
    def from_path(path: GenericPath) -> "StdioDef":
        return StdioDef(path=path, on_close="none")

    @staticmethod
    def remove_if_finished(path: Optional[GenericPath] = None) -> "StdioDef":
        return StdioDef(path=path, on_close="rm-if-finished")


Stdio = Union[GenericPath, StdioDef]


# Keep in sync with `DEFAULT_STDOUT_PATH`
def default_output(ext: str) -> str:
    return "%{CWD}/%{TASK_ID}." + ext


def default_stdout() -> str:
    return default_output("stdout")


def default_stderr() -> str:
    return default_output("stderr")


# TODO: how to resolve TASK_ID in the context of some other task?
class Output:
    def __init__(self, name: str, filepath: Optional[str] = None, extension: Optional[str] = None):
        if filepath and extension:
            raise ValidationException("Parameters `filepath` and `extension` are mutually exclusive")

        self.name = name
        self.filepath = filepath
        self.extension = extension.lstrip(".")


def generate_name(output: Output) -> str:
    if output.filepath is not None:
        return output.filepath
    extension = f".{output.extension}" or ""
    name = generate_random_name()
    return f"{name}{extension}"


def generate_random_name() -> str:
    return uuid.uuid4()[:16]


def materialize_outputs(collection) -> List[Output]:
    items = []
    if isinstance(collection, (list, tuple)):
        items = collection
    elif isinstance(collection, dict):
        items = collection.values()
    elif isinstance(collection, Output):
        items = [collection]
    return [item for item in items if isinstance(item, Output)]


def gather_outputs(collection) -> List[Output]:
    items = []
    if isinstance(collection, (list, tuple)):
        items = collection
    elif isinstance(collection, dict):
        items = collection.values()
    elif isinstance(collection, Output):
        items = [collection]
    return [item for item in items if isinstance(item, Output)]



from pathlib import Path
from typing import Dict, Optional, Sequence, Union

from ..common import GenericPath
from ..ffi import TaskId
from ..ffi.protocol import ResourceRequest
from ..output import Stdio, StdioDef
from ..validation import ValidationException

EnvType = Dict[str, str]


def _make_ffi_requests(
    resources: Optional[Union[ResourceRequest, Sequence[ResourceRequest]]],
):
    if resources is None:
        return ()
    elif isinstance(resources, ResourceRequest):
        return (resources,)
    else:
        return resources


class Task:
    def __init__(
        self,
        task_id: TaskId,
        dependencies: Sequence["Task"] = (),
        priority: int = 0,
        resources: Optional[ResourceRequest] = None,
        env: Optional[EnvType] = None,
        cwd: Optional[GenericPath] = None,
        stdout: Optional[Stdio] = None,
        stderr: Optional[Stdio] = None,
        name: Optional[str] = None,
        crash_limit: Optional[int] = None,
    ):
        assert dependencies is not None
        self.task_id = task_id
        self.dependencies = dependencies
        self.priority = priority
        self.resources = resources
        self.env = env or {}
        self.cwd = str(cwd) if cwd else None
        self.stdout = build_stdio(stdout, "stdout")
        self.stderr = build_stdio(stderr, "stderr")
        self.name = name
        self.crash_limit = crash_limit

    @property
    def label(self) -> str:
        """
        Returns the label of the task.
        If the task has an assigned name, the label is equal to the name.
        Otherwise, the label is the ID of the task converted to a string.
        """
        return self.name if self.name is not None else str(self.task_id)

    def _build(self, client):
        raise NotImplementedError


def build_stdio(stdio: Optional[Stdio], stream: str) -> Optional[StdioDef]:
    if isinstance(stdio, (str, Path)):
        return StdioDef.from_path(stdio)
    elif isinstance(stdio, StdioDef):
        return stdio
    elif stdio is None:
        return None
    else:
        raise ValidationException(
            f"Invalid value provided for `{stream}`: {type(stdio)}. Expected str, Path or `StdioDef` or `None`."
        )



import itertools
import logging
from typing import Dict, Optional, Sequence

from tqdm import tqdm

from .common import GenericPath
from .ffi.client import (
    ClientConnection,
    FailedTaskContext,
    JobId,
    TaskFailureMap,
    TaskId,
)
from .job import HasJobId, Job, SubmittedJob, get_job_id
from .task.function import PythonEnv
from .utils.string import pluralize

MAX_PRINTED_TASKS = 5

JobMap = Dict[JobId, Job]


class FailedJobsException(Exception):
    """
    This exception is triggered if a task fails.
    """

    def __init__(self, failed_tasks: TaskFailureMap, job_map: JobMap):
        self.failed_tasks = failed_tasks
        self.job_map = job_map

    def task_label(self, job_id: JobId, task_id: TaskId) -> str:
        return self.job_map[job_id].task_by_id(task_id).label

    def __str__(self):
        error = ""
        for job_id, tasks in self.failed_tasks.items():
            error += f"The following tasks of job `{job_id}` have failed:\n"
            for task_id, ctx in itertools.islice(tasks.items(), MAX_PRINTED_TASKS):
                task_label = self.task_label(job_id, task_id)
                error += f"Task {task_label} (id={task_id}):\n{ctx.error}\n"
                if ctx.cwd or ctx.stdout or ctx.stderr:
                    error += "You can find more information here:\n"
                    if ctx.cwd:
                        error += f"Working directory: {ctx.cwd}\n"
                    if ctx.stdout:
                        error += f"Stdout: {ctx.stdout}\n"
                    if ctx.stderr:
                        error += f"Stderr: {ctx.stderr}\n"
        return f"{error}\n"


class Client:
    def __init__(
        self,
        server_dir: Optional[GenericPath] = None,
        python_env: Optional[PythonEnv] = None,
    ):
        """
        A client serves as a gateway for submitting jobs and querying information about a running
        HyperQueue server.

        :param server_dir: Path to a server directory of a running HyperQueue server.
        :param python_env: Python environment which configures Python tasks created by
        [`function`](`hyperqueue.job.Job.function`).
        """
        server_dir = str(server_dir) if server_dir else None
        self.connection = ClientConnection(server_dir)
        if python_env is None:
            python_env = PythonEnv()
        self.python_env = python_env

    def submit(self, job: Job) -> SubmittedJob:
        """
        Submit a job into HyperQueue.

        :param job: Job that will be submitted.
        """
        job_desc = job._build(self)
        task_count = len(job_desc.tasks)
        if task_count < 1:
            raise Exception("Submitted job must have at least a single task")

        job_id = self.connection.submit_job(job_desc)
        logging.info(f"Submitted job {job_id} with {task_count} {pluralize('task', task_count)}")
        return SubmittedJob(job=job, id=job_id)

    def wait_for_jobs(self, jobs: Sequence[SubmittedJob], raise_on_error=True) -> bool:
        """Returns True if all tasks were successfully finished"""

        job_ids = tuple(job.id for job in jobs)
        job_ids_str = ",".join(str(id) for id in job_ids)
        if len(jobs) > 1:
            job_ids_str = "{" + job_ids_str + "}"
        logging.info(f"Waiting for {pluralize('job', len(jobs))} {job_ids_str} to finish")

        callback = create_progress_callback()

        failed_jobs = self.connection.wait_for_jobs(job_ids, callback)
        if failed_jobs and raise_on_error:
            failed_tasks = self.connection.get_failed_tasks(failed_jobs)
            job_map = {job.id: job.job for job in jobs}
            raise FailedJobsException(failed_tasks, job_map)
        return len(failed_jobs) == 0

    def get_failed_tasks(self, job: SubmittedJob) -> Dict[TaskId, FailedTaskContext]:
        result = self.connection.get_failed_tasks([job.id])
        return result[job.id]

    def forget(self, job: HasJobId):
        """
        Forget a completed job to free up its resources from the server.

        :param job: Submitted job (or job ID) that will be forgotten.
        """
        job_id = get_job_id(job)
        self.connection.forget_job(job_id)
        logging.info(f"Forgotten job {job_id}")


def create_progress_callback():
    bar = None

    def cb(jobs):
        nonlocal bar

        if bar is None:
            total_tasks = sum(s["total"] for s in jobs.values())
            bar = tqdm(total=total_tasks, unit="task")

        job_count = len(jobs)
        completed = sum(1 if s["completed"] else 0 for s in jobs.values())

        job_text = pluralize("job", job_count)
        bar.set_postfix_str(f"{completed}/{job_count} {job_text} completed")

        finished_tasks = sum(s["finished"] for s in jobs.values())

        delta = finished_tasks - bar.n
        if delta:
            bar.update(delta)

    return cb



from typing import Optional

from . import ffi


class HqClusterContext:
    """
    Opaque class returned from `cluster_start`.
    Should be passed to FFI methods that require it.
    """


class Cluster:
    def __init__(self, directory: Optional[str] = None):
        self.directory = directory
        self.ctx: HqClusterContext = ffi.cluster_start(directory)

    def add_worker(self, cores: int):
        self.ctx.add_worker(cores)

    @property
    def server_dir(self) -> str:
        return self.ctx.server_dir

    def stop(self):
        return self.ctx.stop()



import dataclasses
from typing import Dict, List, Optional, Sequence

from . import JobId, TaskId, ffi
from .protocol import JobDescription


class HqClientContext:
    """
    Opaque class returned from `connect_to_server`.
    Should be passed to FFI methods that require it.
    """


@dataclasses.dataclass(frozen=True)
class FailedTaskContext:
    error: str
    cwd: Optional[str]
    stdout: Optional[str]
    stderr: Optional[str]


TaskFailureMap = Dict[JobId, Dict[TaskId, FailedTaskContext]]


class ClientConnection:
    def __init__(self, directory: Optional[str] = None):
        self.ctx: HqClientContext = ffi.connect_to_server(directory)

    def submit_job(self, job_description: JobDescription) -> JobId:
        return ffi.submit_job(self.ctx, job_description)

    def wait_for_jobs(self, job_ids: Sequence[JobId], callback) -> List[JobId]:
        """Blocks until jobs are finished. Returns the number of failed tasks"""
        return ffi.wait_for_jobs(self.ctx, job_ids, callback)

    def stop_server(self):
        return ffi.stop_server(self.ctx)

    def get_failed_tasks(self, job_ids: Sequence[JobId]) -> TaskFailureMap:
        jobs = ffi.get_failed_tasks(self.ctx, job_ids)
        return {
            job_id: {task_id: FailedTaskContext(**data) for (task_id, data) in task_data.items()}
            for (job_id, task_data) in jobs.items()
        }

    def forget_job(self, job_id: JobId):
        return ffi.forget_job(self.ctx, job_id)



from typing import Dict, List, Optional, Sequence, Union

from ..common import GenericPath
from ..ffi import TaskId
from ..ffi.protocol import ResourceRequest, TaskDescription
from ..output import Output, Stdio, gather_outputs
from ..validation import ValidationException, validate_args
from .task import EnvType, Task, _make_ffi_requests

ProgramArgs = Union[List[str], str]


class ExternalProgram(Task):
    """
    Task that represents the execution of an executable binary.
    """

    def __init__(
        self,
        task_id: TaskId,
        *,
        args: List[str],
        env: Optional[EnvType] = None,
        cwd: Optional[GenericPath] = None,
        stdout: Optional[Stdio] = None,
        stderr: Optional[Stdio] = None,
        stdin: Optional[Union[str, bytes]] = None,
        name: Optional[str] = None,
        dependencies: Sequence[Task] = (),
        task_dir: bool = False,
        priority: int = 0,
        resources: Optional[ResourceRequest],
        crash_limit: Optional[int] = None,
    ):
        super().__init__(
            task_id,
            dependencies,
            priority,
            resources,
            env=env,
            cwd=cwd,
            stdout=stdout,
            stderr=stderr,
            name=name,
            crash_limit=crash_limit,
        )
        args = to_arg_list(args)
        validate_args(args)
        self.args = args
        self.task_dir = task_dir

        if stdin is None or isinstance(stdin, bytes):
            self.stdin = stdin
        elif isinstance(stdin, str):
            self.stdin = stdin.encode()
        else:
            raise Exception("stdin has to be str, bytes, or None")

        self.outputs = get_task_outputs(self)

    def _build(self, client):
        depends_on = [dependency.task_id for dependency in self.dependencies]
        return TaskDescription(
            id=self.task_id,
            args=self.args,
            env=self.env,
            stdout=self.stdout,
            stderr=self.stderr,
            stdin=self.stdin,
            cwd=self.cwd,
            dependencies=depends_on,
            task_dir=self.task_dir,
            priority=self.priority,
            resource_request=_make_ffi_requests(self.resources),
            crash_limit=self.crash_limit,
        )

    def __getitem__(self, key: str):
        if key not in self.outputs:
            raise Exception(f"Output `{key}` not found in {self}")
        return self.outputs[key]

    def __repr__(self):
        return f"Task(args={self.args}, env={self.env}, cwd={self.cwd}, outputs={self.outputs})"


def to_arg_list(args: ProgramArgs) -> List[str]:
    if isinstance(args, str):
        return [args]
    return args


def get_task_outputs(task: ExternalProgram) -> Dict[str, Output]:
    # TODO: outputs in cwd
    # TODO: multiple outputs with the same name, but different parameters
    output_map = {}

    outputs = gather_outputs(task.args) + gather_outputs(task.env)
    for output in outputs:
        if output.name in output_map:
            raise ValidationException(f"Output `{output.name}` has been defined multiple times")
        output_map[output.name] = output
    return output_map



import logging
import os
import pickle
from typing import Callable, Optional, Sequence, Union

from ...common import GenericPath
from ...ffi import TaskId
from ...ffi.protocol import ResourceRequest, TaskDescription
from ..task import EnvType, Task, _make_ffi_requests
from .wrapper import CloudWrapper

_CLOUDWRAPPER_CACHE = {}


def purge_cache():
    global _CLOUDWRAPPER_CACHE
    _CLOUDWRAPPER_CACHE = {}


def cloud_wrap(fn, cache=True) -> CloudWrapper:
    if isinstance(fn, CloudWrapper):
        return fn
    return CloudWrapper(fn, cache=cache)


LOG_LEVELS = {
    "DEBUG": logging.DEBUG,
    "INFO": logging.INFO,
    "WARNING": logging.WARNING,
    "ERROR": logging.ERROR,
}


def get_logging_level(default: int) -> int:
    level = os.environ.get("HQ_PYLOG")
    return LOG_LEVELS.get(level, default)


def task_main():
    import pickle
    import sys

    try:
        logging.basicConfig(
            level=get_logging_level(logging.INFO),
            format="%(asctime)s %(name)s:%(levelname)-4s %(message)s",
            datefmt="%d-%m-%Y %H:%M:%S",
        )

        fn, a, kw = pickle.loads(sys.stdin.buffer.read())
        fn(*a, **(kw if kw is not None else {}))
    except BaseException:
        import os
        import traceback

        t = traceback.format_exc()
        with open(os.environ["HQ_ERROR_FILENAME"], "w") as f:
            f.write(t)
        print(t, file=sys.stderr)
        sys.stdout.flush()
        sys.stderr.flush()
        sys.exit(1)


class PythonEnv:
    def __init__(
        self,
        python_bin: str = "python3",
        prologue: Optional[str] = None,
        shell: str = "bash",
    ):
        """
        Describes an environment for spawning Python interpreters.
        :param python_bin: Python binary that will be executed.
        :param prologue: Shell command that will be executed prior to launching the Python interpreter.
        :param shell: Shell used for executing `prologue`.
        """
        code = "from hyperqueue.task.function import task_main as m; m()"

        if prologue:
            self.args = [shell, "-c", f'{prologue}\n\n{python_bin} -c "{code}"']
        else:
            self.args = [python_bin, "-c", code]


class PythonFunction(Task):
    """
    Task that represents the execution of a Python function.
    """

    def __init__(
        self,
        task_id: TaskId,
        fn: Callable,
        *,
        args=(),
        kwargs=None,
        env: Optional[EnvType] = None,
        cwd: Optional[GenericPath] = None,
        stdout: Optional[GenericPath] = None,
        stderr: Optional[GenericPath] = None,
        name: Optional[str] = None,
        dependencies=(),
        priority: int = 0,
        resources: Optional[Union[ResourceRequest, Sequence[ResourceRequest]]] = None,
        crash_limit: Optional[int] = None,
    ):
        name = generate_task_name(task_id, name, fn)
        super().__init__(
            task_id,
            dependencies,
            priority,
            resources,
            env=env,
            cwd=cwd,
            stdout=stdout,
            stderr=stderr,
            name=name,
            crash_limit=crash_limit,
        )

        fn_id = id(fn)
        wrapper = _CLOUDWRAPPER_CACHE.get(fn_id)
        if wrapper is None:
            wrapper = cloud_wrap(fn)
            _CLOUDWRAPPER_CACHE[fn_id] = wrapper

        self.fn = wrapper
        self.args = args
        self.kwargs = kwargs

    def _build(self, client) -> TaskDescription:
        depends_on = [dependency.task_id for dependency in self.dependencies]
        return TaskDescription(
            id=self.task_id,
            args=client.python_env.args,
            stdout=self.stdout,
            stderr=self.stderr,
            stdin=pickle.dumps((self.fn, self.args, self.kwargs)),
            env=self.env,
            cwd=self.cwd,
            dependencies=depends_on,
            task_dir=True,
            priority=self.priority,
            resource_request=_make_ffi_requests(self.resources),
            crash_limit=self.crash_limit,
        )

    def __repr__(self):
        return f"<PyFunction fn={self.fn.__name__}>"


def generate_task_name(id: TaskId, name: Optional[str], fn: Callable) -> Optional[str]:
    if name is not None:
        return name

    if hasattr(fn, "__name__"):
        return f"{fn.__name__}/{id}"
    return None



import inspect
import logging

import cloudpickle


class CloudWrapper:
    """
    Wraps a callable so that cloudpickle is used to pickle it, caching the pickle.
    """

    def __init__(self, fn, pickled_fn=None, cache=True, protocol=cloudpickle.DEFAULT_PROTOCOL):
        if fn is None:
            if pickled_fn is None:
                raise ValueError("Pass at least one of `fn` and `pickled_fn`")
            fn = cloudpickle.loads(pickled_fn)
        assert callable(fn)
        # Forget pickled_fn if it should not be cached
        if pickled_fn is not None and not cache:
            pickled_fn = None
        if inspect.isasyncgen(fn):
            raise TypeError("async functions not supported")

        self.fn = fn
        self.pickled_fn = pickled_fn
        self.cache = cache
        self.protocol = protocol
        self.__doc__ = "CloudWrapper for {!r}. Original doc:\n\n{}".format(self.fn, self.fn.__doc__)
        if hasattr(self.fn, "__name__"):
            self.__name__ = self.fn.__name__

        # Build-in functions does not have signature
        try:
            self.__signature__ = inspect.signature(self.fn)
        except ValueError:
            pass

    def is_generator_function(self):
        return inspect.isgeneratorfunction(self.fn)

    def __repr__(self):
        return "<{}({!r})>".format(self.__class__.__name__, self.fn)

    def _get_pickled_fn(self):
        "Get cloudpickled version of self.fn, optionally caching the result"
        if self.pickled_fn is not None:
            return self.pickled_fn

        pfn = cloudpickle.dumps(self.fn, protocol=self.protocol)
        if self.cache:
            self.pickled_fn = pfn
        return pfn

    def __call__(self, *args, **kwargs):
        logging.debug(f"Running function {self.fn} using args {args} and kwargs {kwargs}")
        return self.fn(*args, **kwargs)

    def __reduce__(self):
        return (
            self.__class__,
            (None, self._get_pickled_fn(), self.cache, self.protocol),
        )



from .common import GenericPath
from .job import Job
from .task.task import Task
from .utils.package import MissingPackageException


def visualize_job(job: Job, path: GenericPath):
    """
    Visualizes the task graph of the passed job in the DOT format.
    The result is written to a file located at `path`.

    Note: this function requires the `pydot` package to be installed.
    """

    try:
        import pydot
    except ImportError:
        raise MissingPackageException("pydot")

    graph = pydot.Dot("job", graph_type="digraph")
    visited = {}

    def visit(task: Task):
        nonlocal visited, graph
        if task.task_id in visited:
            return visited[task.task_id]

        node = pydot.Node(task.label)
        graph.add_node(node)
        for dep in task.dependencies:
            dep_node = visit(dep)
            edge = pydot.Edge(dep_node.get_name(), node.get_name())
            graph.add_edge(edge)

        visited[task.task_id] = node
        return node

    for task in job.tasks:
        visit(task)

    graph.write(path)



"""
This is the Python API of HyperQueue.

Important classes:

* [`Client`](`hyperqueue.client.Client`) serves for connecting to a HyperQueue server.
* [`LocalCluster`](`hyperqueue.cluster.LocalCluster`) can be used to spawn a local HyperQueue
cluster.
* [`Job`](`hyperqueue.job.Job`) describes a job containing a directed acyclic graph of tasks.
It can be submitted using a client.
"""

# Re-exports
from .client import Client  # noqa: F401
from .cluster import LocalCluster  # noqa: F401
from .ffi import get_version  # noqa: F401
from .job import Job  # noqa: F401



from typing import List


class ValidationException(BaseException):
    pass


def validate_args(args: List[str]):
    from .output import Output

    for arg in args:
        if not isinstance(arg, (str, Output)):
            raise ValidationException(
                "Each program argument must either be a string or an instance of `hq.Output`. "
                f"Argument `{arg}` has type `{type(arg)}`."
            )



import dataclasses
import datetime
from typing import Dict, List, Optional, Sequence, Union

from ..output import StdioDef


class ResourceRequest:
    n_nodes: int = 0
    resources: Dict[str, Union[int, float, str]] = dataclasses.field(default_factory=dict)
    min_time: Optional[datetime.timedelta] = None

    def __init__(
        self,
        *,
        n_nodes=0,
        cpus: Union[int, float, str] = 1,
        resources: Optional[Dict[str, Union[int, float, str]]] = None,
        min_time: Optional[datetime.timedelta] = None,
    ):
        self.n_nodes = n_nodes
        if resources is None:
            resources = {}
        resources["cpus"] = cpus
        self.resources = resources
        self.min_time = min_time

    def __repr__(self):
        return f"<ResourceRequest n_nodes={self.n_nodes} resources={self.resources} min_time={self.min_time}>"


@dataclasses.dataclass()
class TaskDescription:
    id: int
    args: List[str]
    cwd: Optional[str]
    stdout: Optional[StdioDef]
    stderr: Optional[StdioDef]
    stdin: Optional[bytes]
    env: Optional[Dict[str, str]]
    dependencies: Sequence[int]
    task_dir: bool
    priority: int
    resource_request: Sequence[ResourceRequest]
    crash_limit: Optional[int]


@dataclasses.dataclass
class JobDescription:
    name: Optional[str]
    tasks: List[TaskDescription]
    max_fails: Optional[int]



import dataclasses
from pathlib import Path
from typing import Dict, List, Optional, Sequence, Union

from .common import GenericPath
from .ffi import JobId, TaskId
from .ffi.protocol import JobDescription, ResourceRequest
from .output import Stdio, default_stderr, default_stdout
from .task.function import PythonFunction
from .task.program import ExternalProgram, ProgramArgs
from .task.task import EnvType, Task


class Job:
    """
    Represents a HQ job.
    """

    def __init__(
        self,
        default_workdir: Optional[GenericPath] = None,
        max_fails: Optional[int] = 1,
        default_env: Optional[EnvType] = None,
        name: Optional[str] = None,
    ):
        """
        :param default_workdir: Default working directory for tasks.
        :param max_fails: How many tasks can fail before the whole job will be cancelled.
        :param default_env: Environment variables that will be automatically set for each task in
        this job.
        :param name: Name of the job.
        """
        self.tasks: List[Task] = []
        self.task_map: Dict[TaskId, Task] = {}
        self.max_fails = max_fails
        self.default_workdir = Path(default_workdir).resolve() if default_workdir is not None else default_workdir
        self.default_env = default_env or {}
        self.name = name

    def task_by_id(self, id: TaskId) -> Optional[Task]:
        """
        Finds a task with the given ID.
        """
        return self.task_map.get(id)

    def program(
        self,
        args: ProgramArgs,
        *,
        env: Optional[EnvType] = None,
        cwd: Optional[GenericPath] = None,
        stdout: Optional[Stdio] = default_stdout(),
        stderr: Optional[Stdio] = default_stderr(),
        stdin: Optional[Union[str, bytes]] = None,
        deps: Sequence[Task] = (),
        name: Optional[str] = None,
        task_dir: bool = False,
        priority: int = 0,
        resources: Optional[Union[ResourceRequest, Sequence[ResourceRequest]]] = None,
        crash_limit: Optional[int] = None,
    ) -> ExternalProgram:
        """
        Creates a new task that will execute the provided command.

        :param args: List of arguments will be executed. The arguments have to be strings.
        :param env: Environment variables passed to the executed command.
        :param cwd: Working directory of the executed command.
        :param stdout: Path to a file that will store the standard output of the executed command.
        :param stderr: Path to a file that will store the standard error output of the executed
        command.
        :param stdin: If provided, these bytes will be passed as the standard input of the executed
        command.
        :param deps: A sequence of dependencies that have to be completed first before this task
        can start executing.
        :param name: Name of the task.
        :param task_dir: If True, an isolated directory will be created for the task.
        :param priority: Priority of the created task.
        :param resources: List of resource requests required by this task.
        :param crash_limit: How many times can a worker that executes this task crash before the
        task is considered to be failed.
        """
        task = ExternalProgram(
            len(self.tasks),
            args=args,
            env=merge_envs(self.default_env, env),
            cwd=cwd or self.default_workdir,
            dependencies=deps,
            stdout=stdout,
            stderr=stderr,
            stdin=stdin,
            name=name,
            task_dir=task_dir,
            priority=priority,
            resources=resources,
            crash_limit=crash_limit,
        )
        self._add_task(task)
        return task

    def function(
        self,
        fn,
        *,
        args=(),
        kwargs=None,
        env: Optional[EnvType] = None,
        cwd: Optional[GenericPath] = None,
        stdout: Optional[Stdio] = default_stdout(),
        stderr: Optional[Stdio] = default_stderr(),
        deps: Sequence[Task] = (),
        name: Optional[str] = None,
        priority: int = 0,
        resources: Optional[Union[ResourceRequest, Sequence[ResourceRequest]]] = None,
        crash_limit: Optional[int] = None,
    ) -> PythonFunction:
        """
        Creates a new task that will execute the provided Python function.

        :param args: Positional arguments that will be passed to the Python function.
        :param kwargs: Keyword arguments that will be passed to the Python function.
        :param env: Environment variables passed to the executed command.
        :param cwd: Working directory of the executed command.
        :param stdout: Path to a file that will store the standard output of the executed command.
        :param stderr: Path to a file that will store the standard error output of the executed
        command.
        :param deps: A sequence of dependencies that have to be completed first before this task
        can start executing.
        :param name: Name of the task.
        :param priority: Priority of the created task.
        :param resources: List of resource requests required by this task.
        :param crash_limit: How many times can a worker that executes this task crash before the
        task is considered to be failed.
        """
        task = PythonFunction(
            len(self.tasks),
            fn,
            args=args,
            kwargs=kwargs,
            env=merge_envs(self.default_env, env),
            cwd=cwd or self.default_workdir,
            stdout=stdout,
            stderr=stderr,
            name=name,
            dependencies=deps,
            priority=priority,
            resources=resources,
            crash_limit=crash_limit,
        )
        self._add_task(task)
        return task

    def _add_task(self, task: Task):
        self.tasks.append(task)
        self.task_map[task.task_id] = task

    def _build(self, client) -> JobDescription:
        task_descriptions = []
        for task in self.tasks:
            task_descriptions.append(task._build(client))
        return JobDescription(name=self.name, tasks=task_descriptions, max_fails=self.max_fails)


@dataclasses.dataclass
class SubmittedJob:
    """
    Successfully submitted job.
    """

    job: Job
    id: JobId


HasJobId = Union[JobId, SubmittedJob]


def get_job_id(job: HasJobId) -> JobId:
    if isinstance(job, SubmittedJob):
        return job.id
    elif isinstance(job, (JobId, int)):
        return job
    else:
        raise ValueError(f"Invalid value of type {type(job)} passed to `get_job_id`")


def merge_envs(default: EnvType, env: Optional[EnvType]) -> EnvType:
    environment = default.copy()
    if env:
        environment.update(env)
    return environment



class MissingPackageException(BaseException):
    def __init__(self, package: str):
        self.package = package

    def __str__(self):
        return f"Unable to import `{self.package}`. You have to install the `{self.package}` package."



import dataclasses
import multiprocessing
from pathlib import Path
from typing import Optional

from ..client import Client
from ..ffi.cluster import Cluster


@dataclasses.dataclass
class WorkerConfig:
    """
    Configuration of a worker spawned by a local cluster.
    """

    # If None, use all available cores
    cores: Optional[int] = None


class LocalCluster:
    """
    Represents a local deployed HyperQueue infrastructure.

    You can use `LocalCluster` to quickly spin up a HyperQueue server along with a set of workers
    locally.

    The cluster can be used as a context manager. It will be stopped when the context ends:
    ```python
    with LocalCluster() as cluster:
        client = cluster.client()
        ...
    # The cluster was stopped
    ```
    """

    def __init__(
        self,
        server_dir: Optional[Path] = None,
        worker_config: Optional[WorkerConfig] = None,
    ):
        """
        :param server_dir: Server directory where will the cluster store its files.
        :param worker_config: Configuration of workers spawned in the cluster.
        """
        self.cluster = Cluster(server_dir)
        if worker_config is not None:
            self.start_worker(worker_config)

    def start_worker(self, config: WorkerConfig = None):
        """
        Adds a new worker with the given `config` to the cluster.
        """
        config = config if config is not None else WorkerConfig()
        cores = config.cores or multiprocessing.cpu_count()
        self.cluster.add_worker(cores)

    def client(self, **client_args) -> Client:
        """
        Creates a client connected to this cluster.
        """
        return Client(self.cluster.server_dir, **client_args)

    def stop(self):
        """
        Stops the server and all workers of this cluster.
        """
        self.cluster.stop()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()
